@@ -1,8 +1,23 @@
 from __future__ import annotations
 
 from itertools import islice
-from typing import TYPE_CHECKING, Any, Callable, Collection, Generator, Iterable, List, Optional, Sequence, Tuple, cast
+from typing import (
+    TYPE_CHECKING,
+    Any,
+    Callable,
+    Collection,
+    Dict,
+    Generator,
+    Iterable,
+    List,
+    Optional,
+    Sequence,
+    Tuple,
+    Union,
+    cast,
+)
 
+from ..scripture.verse_ref import VerseRef
 from ..tokenization.detokenizer import Detokenizer
 from ..tokenization.tokenizer import Tokenizer
 from .aligned_word_pair import AlignedWordPair
@@ -13,6 +28,10 @@ from .token_processors import escape_spaces, lowercase, normalize, unescape_spac
 
 if TYPE_CHECKING:
     import pandas as pd
+    from datasets.arrow_dataset import Dataset
+    from datasets.info import DatasetInfo
+    from datasets.iterable_dataset import IterableDataset
+    from datasets.splits import NamedSplit
 
 
 class ParallelTextCorpus(Corpus[ParallelTextRow]):
@@ -25,8 +44,44 @@ class ParallelTextCorpus(Corpus[ParallelTextRow]):
         source_column: str = "source",
         target_column: str = "target",
         alignment_column: Optional[str] = "alignment",
+        default_text_id: str = "*all*",
+        ref_factory: Optional[Callable[[Any], Any]] = None,
     ) -> ParallelTextCorpus:
-        return _PandasParallelTextCorpus(df, text_id_column, ref_column, source_column, target_column, alignment_column)
+        return _PandasParallelTextCorpus(
+            df,
+            text_id_column,
+            ref_column,
+            source_column,
+            target_column,
+            alignment_column,
+            default_text_id,
+            ref_factory,
+        )
+
+    @classmethod
+    def from_hf_dataset(
+        cls,
+        ds: Union[Dataset, IterableDataset],
+        source_lang: str,
+        target_lang: str,
+        text_id_column: Optional[str] = "text",
+        ref_column: Optional[str] = "ref",
+        translation_column: str = "translation",
+        alignment_column: Optional[str] = "alignment",
+        default_text_id: str = "*all*",
+        ref_factory: Optional[Callable[[Any], Any]] = None,
+    ) -> ParallelTextCorpus:
+        return _DatasetParallelTextCorpus(
+            ds,
+            source_lang,
+            target_lang,
+            text_id_column,
+            ref_column,
+            translation_column,
+            alignment_column,
+            default_text_id,
+            ref_factory,
+        )
 
     def invert(self) -> ParallelTextCorpus:
         def _invert(row: ParallelTextRow) -> ParallelTextRow:
@@ -238,7 +293,7 @@ class ParallelTextCorpus(Corpus[ParallelTextRow]):
         refs: Optional[List[Any]] = None if ref_column is None else []
         source: Optional[List[str]] = None if source_column is None else []
         target: Optional[List[str]] = None if target_column is None else []
-        alignments: Optional[List[str]] = None if alignment_column is None else []
+        alignments: Optional[List[List[Tuple[int, int]]]] = None if alignment_column is None else []
         has_alignments = False
         with self.get_rows() as rows:
             for row in rows:
@@ -251,10 +306,11 @@ class ParallelTextCorpus(Corpus[ParallelTextRow]):
                 if target is not None:
                     target.append(row.target_text)
                 if alignments is not None:
-                    alignment = ""
                     if row.aligned_word_pairs is not None:
-                        alignment = " ".join(f"{r.source_index}-{r.target_index}" for r in row.aligned_word_pairs)
+                        alignment = [(r.source_index, r.target_index) for r in row.aligned_word_pairs]
                         has_alignments = True
+                    else:
+                        alignment = []
                     alignments.append(alignment)
 
         data = {}
@@ -270,6 +326,66 @@ class ParallelTextCorpus(Corpus[ParallelTextRow]):
             data[alignment_column] = alignments
         return pd.DataFrame(data)
 
+    def to_hf_dataset(
+        self,
+        source_lang: str,
+        target_lang: str,
+        text_id_column: Optional[str] = "text",
+        ref_column: Optional[str] = "ref",
+        translation_column: str = "translation",
+        alignment_column: Optional[str] = "alignment",
+        info: Optional[DatasetInfo] = None,
+        split: Optional[NamedSplit] = None,
+        format_type: Optional[str] = None,
+    ) -> IterableDataset:
+        try:
+            from datasets.features.features import Features, FeatureType, Sequence, Value
+            from datasets.features.translation import Translation
+            from datasets.info import DatasetInfo
+            from datasets.iterable_dataset import iterable_dataset
+        except ImportError:
+            raise RuntimeError("datasets is not installed.")
+
+        features_dict: Dict[str, FeatureType] = {translation_column: Translation(languages=[source_lang, target_lang])}
+        if text_id_column is not None:
+            features_dict[text_id_column] = Value("string")
+        if ref_column is not None:
+            features_dict[ref_column] = Sequence(Value("string"))
+        if alignment_column is not None:
+            features_dict[alignment_column] = Sequence({source_lang: Value("int32"), target_lang: Value("int32")})
+        features = Features(features_dict)
+
+        if info is not None and info.features is not None and info.features != features:
+            raise ValueError("Features specified in `info.features` is not compatible.")
+
+        if info is None:
+            info = DatasetInfo()
+
+        info.features = features
+
+        def iterable() -> Iterable[Tuple[Union[str, int], dict]]:
+            with self.get_rows() as rows:
+                for row in rows:
+                    key = row.ref
+                    if not isinstance(key, int) and not isinstance(key, str):
+                        key = str(key)
+                    example = {}
+                    if text_id_column is not None:
+                        example[text_id_column] = row.text_id
+                    if ref_column is not None:
+                        example[ref_column] = row.refs
+                    example[translation_column] = {source_lang: row.source_text, target_lang: row.target_text}
+                    if alignment_column is not None and row.aligned_word_pairs is not None:
+                        src_indices: List[int] = []
+                        trg_indices: List[int] = []
+                        for wp in row.aligned_word_pairs:
+                            src_indices.append(wp.source_index)
+                            trg_indices.append(wp.target_index)
+                        example[alignment_column] = {source_lang: src_indices, target_lang: trg_indices}
+                    yield key, example
+
+        return iterable_dataset(iterable(), info, split, format_type)
+
 
 def flatten_parallel_text_corpora(corpora: Iterable[ParallelTextCorpus]) -> ParallelTextCorpus:
     corpus_list = list(corpora)
@@ -342,6 +458,8 @@ class _PandasParallelTextCorpus(ParallelTextCorpus):
         source_column: str,
         target_column: str,
         alignment_column: Optional[str],
+        default_text_id: str,
+        ref_factory: Optional[Callable[[Any], Any]],
     ) -> None:
         self._df = df
         self._text_id_column = text_id_column
@@ -349,6 +467,8 @@ class _PandasParallelTextCorpus(ParallelTextCorpus):
         self._source_column = source_column
         self._target_column = target_column
         self._alignment_column = alignment_column
+        self._default_text_id = default_text_id
+        self._ref_factory = ref_factory
 
     @property
     def missing_rows_allowed(self) -> bool:
@@ -361,24 +481,121 @@ class _PandasParallelTextCorpus(ParallelTextCorpus):
 
     def _get_rows(self) -> Generator[ParallelTextRow, None, None]:
         for index, row in self._df.iterrows():
-            text_id = "*all*"
+            text_id = self._default_text_id
             if self._text_id_column is not None and self._text_id_column in self._df:
                 text_id = cast(str, row[self._text_id_column])
-            ref = index
+            refs = [index]
             if self._ref_column is not None and self._ref_column in self._df:
-                ref = row[self._ref_column]
+                refs = row[self._ref_column]
+                if not isinstance(refs, list):
+                    refs = [refs]
+                if self._ref_factory is not None:
+                    refs = [self._ref_factory(ref) for ref in refs]
             source = cast(str, row[self._source_column])
             target = cast(str, row[self._target_column])
             alignment: Optional[Collection[AlignedWordPair]] = None
             if self._alignment_column is not None and self._alignment_column in self._df:
                 v = row[self._alignment_column]
-                alignment = AlignedWordPair.parse(v) if isinstance(v, str) else v
+                alignment = AlignedWordPair.parse(v) if isinstance(v, str) else [AlignedWordPair(t[0], t[1]) for t in v]
             yield ParallelTextRow(
                 text_id,
-                [ref],
-                [ref],
+                refs,
+                refs,
                 [source] if len(source) > 0 else [],
                 [target] if len(target) > 0 else [],
                 alignment,
                 is_empty=len(source) == 0 or len(target) == 0,
             )
+
+
+class _DatasetParallelTextCorpus(ParallelTextCorpus):
+    def __init__(
+        self,
+        ds: Union[Dataset, IterableDataset],
+        source_lang: str,
+        target_lang: str,
+        text_id_column: Optional[str],
+        ref_column: Optional[str],
+        translation_column: str,
+        alignment_column: Optional[str],
+        default_text_id: str,
+        ref_factory: Optional[Callable[[Any], Any]],
+    ) -> None:
+        self._ds = ds
+        self._source_lang = source_lang
+        self._target_lang = target_lang
+        self._text_id_column = text_id_column
+        self._ref_column = ref_column
+        self._translation_column = translation_column
+        self._alignment_column = alignment_column
+        self._default_text_id = default_text_id
+        self._ref_factory = ref_factory
+
+    @property
+    def missing_rows_allowed(self) -> bool:
+        return False
+
+    def count(self, include_empty: bool = True) -> int:
+        try:
+            from datasets.arrow_dataset import Dataset
+        except ImportError:
+            raise RuntimeError("datasets is not installed.")
+        if include_empty and isinstance(self._ds, Dataset):
+            return len(self._ds)
+
+        count = 0
+        for example in cast(Iterable[dict], self._ds):
+            if (
+                self._get_translation(self._source_lang, example) != ""
+                and self._get_translation(self._target_lang, example) != ""
+            ):
+                count += 1
+        return count
+
+    def _get_rows(self) -> Generator[ParallelTextRow, None, None]:
+        index = 0
+        for example in cast(Iterable[dict], self._ds):
+            text_id = self._default_text_id
+            refs = [index]
+            if self._ref_column is not None and self._ref_column in example:
+                refs = example[self._ref_column]
+                if not isinstance(refs, list):
+                    refs = [refs]
+                if self._ref_factory is not None:
+                    refs = [self._ref_factory(ref) for ref in refs]
+                    if len(refs) > 0 and isinstance(refs[0], VerseRef):
+                        text_id = refs[0].book
+            if self._text_id_column is not None and self._text_id_column in example:
+                text_id = cast(str, example[self._text_id_column])
+            source = self._get_translation(self._source_lang, example)
+            target = self._get_translation(self._target_lang, example)
+            alignment: Optional[Collection[AlignedWordPair]] = None
+            if self._alignment_column is not None and self._alignment_column in example:
+                src_indices = example[self._alignment_column][self._source_lang]
+                trg_indices = example[self._alignment_column][self._target_lang]
+                alignment = [AlignedWordPair(si, ti) for (si, ti) in zip(src_indices, trg_indices)]
+
+            yield ParallelTextRow(
+                text_id,
+                refs,
+                refs,
+                [source] if len(source) > 0 else [],
+                [target] if len(target) > 0 else [],
+                alignment,
+                is_empty=len(source) == 0 or len(target) == 0,
+            )
+            index += 1
+
+    def _get_translation(self, lang: str, example: dict) -> str:
+        translation: dict = example[self._translation_column]
+        if lang in translation:
+            return translation[lang]
+        if "language" in translation and "translation" in translation:
+            languages: List[str] = translation["language"]
+            translations: List[str] = translation["translation"]
+            try:
+                index = languages.index(lang)
+                return translations[index]
+            except ValueError:
+                return ""
+        return ""
