@@ -1,13 +1,19 @@
+from typing import IO
+
 from chardet import UniversalDetector
 
 from .typeshed import StrPath
 
 
 def detect_encoding(filename: StrPath) -> str:
-    detector = UniversalDetector()
     with open(filename, "rb") as file:
-        for line in file:
-            detector.feed(line)
-            if detector.done:
-                break
+        return detect_encoding_from_stream(file)
+
+
+def detect_encoding_from_stream(stream: IO[bytes]) -> str:
+    detector = UniversalDetector()
+    for line in stream:
+        detector.feed(line)
+        if detector.done:
+            break
     return detector.close()["encoding"]
